if __name__ == "__main__":
    from .reconciliation import reconciliate_many, recon_table_to_str
    from .error import InconsistentTrees, MissedData, ParameterError
    from .orthology import orthologs_from_trees_df
    from .supertree import from_gene_forest
    from .constants import _df_matches_cols, _default_f
    from .hug_free import build_graph_series
    from .parse_prt import _parse, normalization_modes, normalization_smallest
    from .common_tools import norm_path
    from .cBMG_tools import analyze_digraphs_list
    from .nhxx_tools import read_nhxx, get_nhx
    from .in_out import read_tl_digraph, tl_digraph_from_pandas

    from tqdm import tqdm
    from numpy import argsort
    import pandas as pd
    import os

    import argparse
    parser = argparse.ArgumentParser(prog= 'revolutionhtl',
                                     description='Reconstruction of Evolutionaty Histories TooL.\n1. Best hit selection 2. Orthology and gene tree reconstruction 3. Species tree reconstruction 4. Tree reconciliation\nhttps://pypi.org/project/revolutionhtl/',
                                     usage='python -m revolutionhtl <arguments>',
                                     formatter_class=argparse.MetavarTypeHelpFormatter,
                                    )

    # Arguments
    ###########

    # Input data
    # ..........

    parser.add_argument('-steps',
                        help= 'List of steps to run (default: 1 2 3 4).',
                        type= int,
                        nargs= '*',
                        default= [1, 2, 3, 4]
                       )

    parser.add_argument('-alignment_h', '--alignment_hits',
                        help= 'Directory containing alignment hits, the input of step 1. (default: ./).',
                        type= str,
                        default= './'
                       )

    parser.add_argument('-best_h', '--best_hits',
                        help= '.tsv file containing best hits, the input of step 2. (default: use output of step 1).',
                        type= str,
                       )

    parser.add_argument('-T', '--gene_trees',
                        help= '.tsv file containing gene trees, the input of steps 3 and 4. (default: use output of step 2).',
                        type= str,
                       )

    parser.add_argument('-S', '--species_tree',
                        help= '.nhx file containing a species tree, an input of step 4. (default: use output of step 3).',
                        type= str,
                        default= 'BUILD_FROM_GENE_FOREST',
                       )

    # Parameters
    # ..........
    parser.add_argument('-bh_heuristic', '--besthit_heuristic',
                        help= 'string, indicates how to normalize bit-score in step 1 (default: normal). Normal: no normalization, prt: use proteinortho auxiliary files, smallest: use length of the smallest sequence, target: use target sequence, query: use query sequence, directed: x->y hit, bidirectional: use x->y and y->x hits..',
                        type= str,
                        choices= ['normal', 'prt']+normalization_smallest+normalization_modes,
                        default= 'normal',
                       )

    parser.add_argument('-f', '--f_value',
                        help= f'Real number between 0 and 1, a parameter of step 1. Defines the adaptative threshhold as: f*max_bit_score (default: {_default_f}).',
                        type= float,
                        default= _default_f,
                       )

    parser.add_argument('-bmg_h', '--bmg_heuristic',
                        help= f'Comunity detection method, an heuristic of step 2. (default: Louvain).',
                        type= str,
                        default= 'Louvain',
                        choices= ['Mincut', 'BPMF', 'Karger', 'Greedy', 'Gradient_Walk', 'Louvain', 'Louvain_Obj'],
                       )

    parser.add_argument('-no_binary_R', '--no_binary_triples',
                        help= 'Flag, specifies if use binary triples in step 2. (no flag: use binary triples, flag: do not use binary triples).',
                        action= 'store_false',
                       )


    parser.add_argument('-f_bT', '--force_binary_gene_tree',
                        help= 'Flag, specifies if force binary tree in step 2. (no flag: do not force binary, flag: force binary).',
                        action= 'store_true',
                       )

    parser.add_argument('-T_no_db', '--gene_tree_no_double_build',
                        help= 'Flag, specifies if run build algorithm twice to obtain less resolved tree in step 2. (no flag: double build, flag: single build).',
                        action= 'store_false',
                       )

    parser.add_argument('-stree_h', '--species_tree_heuristic',
                        help= 'Comunity detection method, an heuristic of step 3. (default: louvain_weight).',
                        type= str,
                        default= 'louvain_weight',
                        choices= ['naive', 'louvain', 'mincut', 'louvain_weight']
                       )

    parser.add_argument('-streeh_repeats', '--stree_heuristic_repeats',
                        help= 'integer, specifies how many times run the heuristic of step 3. (default: 3)',
                        type= int,
                        default= 3
                       )

    parser.add_argument('-streeh_b', '--streeh_binary',
                        help= 'Flag, specifies if force binary tree in step 3. (no flag: do not force binary, flag: force binary).',
                        action= 'store_true',
                       )

    parser.add_argument('-streeh_ndb', '--streeh_no_doble_build',
                        help= 'Flag, specifies if run build algorithm twice to obtain less resolved tree in step 3. (no flag: double build, flag: single build).',
                        action= 'store_false',
                       )

    # Format parameters
    # .................

    parser.add_argument('-o', '--output_prefix',
                        help= 'Prefix used for output files (default "tl_project").',
                        type= str,
                        default= 'tl_project',
                       )

    parser.add_argument('-og', '--orthogroup_column',
                        help= 'Column in -best_h, -T, and output files specifying orthogroups (default: OG).',
                        type= str,
                        default= 'OG',
                       )

    parser.add_argument('-Nm', '--N_max',
                        help= 'Indicates the maximum number of genes in a orthogroup, bigger orthogroups are splitted. If 0, no orthogroup is splitted. (default= 2000).',
                        type= int,
                        default= 2000,
                       )

    parser.add_argument('-k', '--k_size_partition',
                        help= 'Integer indicatng how many best hit graphs will be processed in bunch:: first graphs with <k genes, then <2k. then <3k, and so on. (default: k=100)',
                        type= int,
                        default= 100,
                       )

    parser.add_argument('-S_attr', '--S_attr_sep',
                        help= 'String used to separate attributes in the species tree at step 4.',
                        type= str,
                        default= ';',
                       )

    args= parser.parse_args()

    ################
    # Process data #
    ################

    from .hello import hello5
    print(hello5)

    allowed_steps= {1, 2, 3, 4}
    bad_steps= set(args.steps) - allowed_steps
    if len(bad_steps) > 0:
        raise ParameterError(f'Only steps 1, 2, and 3 are allowed to be used in the parameter -steps. ')
    else:
        args.steps= sorted(set(args.steps))

    print(f'Running steps {", ".join(map(str, args.steps))}')

    # 1. Convert proteinortho output to a best-hit list
    #################################################
    if 1 in args.steps:
        print("\nStep 1: Convert proteinortho output to a best-hit list")
        print("------------------------------------------------------")
        if args.alignment_hits==None:
            raise MissedData('Step 0 needs a value for the parameter -alignment_hits.')

        args.alignment_hits= norm_path(args.alignment_hits)
        opath= args.output_prefix+'.best_hits.tsv'
        df_hits, OGs_table= _parse(args.alignment_hits, args.f_value, opath, args.besthit_heuristic, N_max= args.N_max)
        print('This file will be used as input for step 2.')

        print('Writing orthogroups...')
        opath= args.output_prefix+'.orthogroups.tsv'
        OGs_table.to_csv(opath, sep='\t', index= False)
        print(f'Successfully written to {opath}')
    else:
        df_hits= None

    # 2. Conver best hit graphs to cBMGs and gene trees
    ##################################################
    if 2 in args.steps:
        print("\nStep 2: Conver best-hit graphs to cBMGs and gene trees")
        print("------------------------------------------------------")
        readHs= type(df_hits)!=pd.DataFrame
        if readHs and (args.best_hits==None):
            raise MissedData('Step 2 needs a value for the parameter -best_hits. Create it by running step 1.')

        if readHs:
            print('Reading hit graphs...')
            G= read_tl_digraph(args.best_hits, og_col= args.orthogroup_column)
        else:
            print('Creating graphs...')
            G= tl_digraph_from_pandas(df_hits, og_col= args.orthogroup_column)

        Tg= build_graph_series(G, args)

        # Print orthologs
        TTg= Tg.reset_index()
        df_orthologs= orthologs_from_trees_df(TTg,
                                forbidden_leaf_name= 'X', tree_col= 'tree')
        opath= f'{args.output_prefix}.orthologs.tsv'
        df_orthologs.to_csv(opath, sep= '\t', index= False)
        print(f'Orthologs successfully written to {opath}')


    else:
        Tg= None

    # 3. Reconstruct species tree
    #############################
    if 3 in args.steps:
        print("\nStep 3: Species tree reconstruction")
        print("-----------------------------------")
        readTg= type(Tg)!=pd.Series
        if readTg and (args.gene_trees==None):
            raise MissedData('Step 3 needs a value for the parameter -gene_trees. Create it by running step 2.')
        if readTg:
            print('Reading trees...')
            gTrees= pd.read_csv(args.gene_trees,
                                sep= '\t').set_index(
                args.orthogroup_column).tree.apply(lambda x: read_nhxx(x, name_attr= 'accession'))
        else:
            gTrees= Tg

        print("Reconstructing species tree...")
        species_tree= from_gene_forest(gTrees,
                                       method= args.species_tree_heuristic,
                                       numb_repeats= args.stree_heuristic_repeats,
                                       doble_build= args.streeh_no_doble_build,
                                       binary= args.streeh_binary,
                                      )

        print("Writing species tree...")
        s_newick= get_nhx(species_tree, root= 1, name_attr= 'species', ignore_inner_name= True)
        opath= f'{args.output_prefix}.species_tree.tsv'
        with open(opath, 'w') as F:
            F.write(s_newick)
        print(f'Successfully written to {opath}')
    else:
        species_tree= None

    # 4. Reconciliate gene trees and species tree
    ##############################################
    if 4 in args.steps:
        print("\nStep 4: Reconciliation of gene species trees")
        print("--------------------------------------------")
        readTg= type(Tg)!=pd.Series
        if readTg:
            if args.gene_trees==None:
                raise MissedData('Step 4 needs a value for the parameter -gene_trees. You can create it by running step 2.')
            print('Reading trees...')
            gTrees= pd.read_csv(args.gene_trees,
                                sep= '\t').set_index(
                args.orthogroup_column).tree.apply(lambda x: read_nhxx(x, name_attr= 'accession'))
        else:
            gTrees= Tg

        readTs= species_tree == None
        if readTs:
            if args.species_tree=='BUILD_FROM_GENE_FOREST':
                raise MissedData('Step 4 needs a species tree, run step 3 to construct one or set a value for the parameter -species_tree')
            print('Reading species tree...')
            with open(args.species_tree) as F:
                sTree= read_nhxx(''.join( F.read().strip().split('\n') ),
                                 name_attr= 'species',
                                 attr_sep= args.S_attr_sep
                                )
        else: 
            sTree= species_tree

        print('Reconciling trees...')
        df_recs= reconciliate_many(gTrees, sTree)

        # Write resolved trees
        df_r= recon_table_to_str(sTree, df_recs, args.orthogroup_column)
        print("Writing to file...")
        opath= f'{args.output_prefix}.reconciliation.tsv'
        df_r.to_csv(opath, sep= '\t', index= False)
        print(f'Reconciliation were successfully written to {opath}')

        # Write labeled species tree
        nhx_s= get_nhx(sTree, name_attr= 'species', root= 1, ignore_inner_name= True)
        opath= f'{args.output_prefix}.labeled_species_tree.nhxx'
        with open(opath, 'w') as F:
            F.write(nhx_s)
        print(f'Indexed species tree successfully written to {opath}')

        """
        # Write orthology relations
        df_orthologs, df_pseudo= orthologs_from_trees_df(df_recs, forbidden_leaf_name= 'X', tree_col= 'reconciliated_tree')

        opath= f'{args.output_prefix}.orthologs.tsv'
        df_orthologs.to_csv(opath, sep= '\t', index= False)
        print(f'Orthologs successfully written to {opath}')

        opath= f'{args.output_prefix}.pseudo_orthologs.tsv'
        df_pseudo.to_csv(opath, sep= '\t', index= False)
        print(f'Pseudo-orthologs successfully written to {opath}')
        """

    print("\nREvolutionH-tl finished all the tasks without any problem")
    print("---------------------------------------------------------")



import pandas as pd
import os
import numpy as np
import networkx as nx
from itertools import chain
from tqdm import tqdm
tqdm.pandas()

from .common_tools import norm_path
from .in_out import create_cBMG, tl_digraph_from_pandas
from .reduce_giant import reduce_graph

####################
#                  #
# Some parameters  #
#                  #
####################

from .constants import _diamond_tl_headers, _default_f, _df_matches_cols

####################
#                  #
# Parse functions  #
#                  #
####################


normalization_modes= ['query_directed', 
                      'target_directed', 
                      'alignment_directed', 
                      'query_bidirectional', 
                      'target_bidirectional', 
                      'alignment_bidirectional',
                     ]
normalization_smallest=['smallest_bidirectional', 'smallest_directed']

colRelation= {'query' : 'Query_length',
              'target' : 'Target_length',
              'alignment' : 'Alignment_ength',
              'smallest' : None,
             }

def _parse(path, f_value, opath, mode= 'normal', N_max= 2000):
    if mode=='prt':
        df_hits= parse_prt_files(path, None,
                                 pd_params= dict(names= _df_matches_cols, sep= '\t'),
                                 species_parser= prt_species_parser
                                )
    elif mode=='normal':
        df_hits= parse_prt_files(path, None)
    elif mode in normalization_modes+normalization_smallest:
        df_hits= parse_prt_files(path, mode)
    else:
        raise ValueError('Mode for parsing only can be one of: "normal", "proteinortho"')
    # select best hits
    df= select_best_hits(df_hits, f_value)
    # Identify orthogroups
    df, OGs_table= identify_orthogroups(df, N_max= N_max)
    df.reset_index(drop=True).to_csv(opath, sep='\t', index= False)
    print(f'Best hits were successfully written to {opath}')
    return df, OGs_table

def simple_species_parser(file):
    species= file.split('/')[-1].split('.')
    return species[0], species[2]

def prt_species_parser(file):
    species= file.split('/')[-1].split('.')
    return species[0], species[3]

def select_best_hits(df_hits, f_value, a= 'Query_accession', b= 'Target_accession'):
    """
    Input:
    - path: A directory containing reciprocal blast-like analisis between pairs of species.
            The files contained there should be named as follows: <species_a>.fa.vs<species_b.fa>.blast
    - f_value: used for the dinamic threshold.
    """
    # Identify best hit scores w.r.t. query_gene and target_species
    w_x_B= max_hit_scores(df_hits)
    # Apply dynamic threshold for best hit selection
    print('Selecting best hits by dynamic threshold...')
    is_best_hit= lambda row: row.Bit_score >= f_value * w_x_B[ row.Query_accession, row.Target_species ]
    mask= df_hits.progress_apply(is_best_hit, axis= 1)
    df_best_hits= df_hits.loc[mask]
    return df_best_hits

def identify_orthogroups(df_best_hits,
                         Query_accession= 'Query_accession',
                         Target_accession= 'Target_accession',
                         Query_species= 'Query_species',
                         Target_species= 'Target_species',
                         N_max= 2000):
    # Create graph
    G= nx.DiGraph()
    G.add_edges_from( df_best_hits[[Query_accession,Target_accession]].values )
    # Identify connected components
    CCs= pd.Series(list(nx.weakly_connected_components(G)))
    # Reduce giant connected components
    mask_giant= CCs.apply(lambda X: len(X)>N_max)
    reduced_CCs= CCs[ mask_giant ].apply(lambda X: reduce_graph(nx.induced_subgraph(G, X), N_max= N_max))
    # Obtain singletones list
    singletones= reduced_CCs.str[1]
    singletones= set(chain.from_iterable(singletones.values))
    # Obtain graph partitions
    reduced_CCs= reduced_CCs.str[0]
    reduced_CCs= pd.Series(list(chain.from_iterable(reduced_CCs.values)))
    # Make final CCs list
    CCs= pd.concat((CCs[ ~mask_giant ], reduced_CCs), ignore_index= True)
    # Sort connected components by size and gene names
    CCs= pd.DataFrame(dict(CCs= CCs,
                           N= CCs.str.len(),
                           name= CCs.apply(lambda X: ''.join(sorted(''.join(X)))),
                          ))
    CCs= CCs.sort_values(by= ['N', 'name']).CCs
    # Asign a number for each connected component
    idx_2_og= dict(enumerate(CCs.values))
    D= {x: i for i,X in idx_2_og.items() for x in X}
    # Quit singletones w.r.t partition
    F= lambda x: x not in singletones
    df_best_hits= df_best_hits[ df_best_hits[Query_accession].apply(F) & df_best_hits[Target_accession].apply(F) ]
    # Keep only intra-hits w.r.t partition
    df_best_hits= df_best_hits[ df_best_hits.apply(lambda X: D[X[Query_accession]]==D[X[Target_accession]], axis= 1) ]
    # Asign a connected component to each row of dataframe
    cols= list(df_best_hits.columns)
    df_best_hits['OG']= df_best_hits[Query_accession].apply(lambda x: D[x])
    # Sort rows by OG
    df_best_hits= df_best_hits[['OG']+cols].sort_values('OG')

    # Create OGs table
    OGs_table= create_og_table(df_best_hits, idx_2_og,
                               Query_accession,
                               Target_accession,
                               Query_species,
                               Target_species)

    return df_best_hits, OGs_table

def get_species_columns(genes, species, gene2species, idx):
    ret= {}
    for gene in genes:
        spec= gene2species[gene]
        ret[spec]= ret.get(spec, []) + [gene]

    nGenes= len(genes)
    nSpecies= len(ret)

    return [idx, nGenes, nSpecies] + [','.join(ret.get(spec, []))
                                      for spec in species]

def create_og_table(df_BHs, idx_2_og,
                    Query_accession,
                    Target_accession,
                    Query_species,
                    Target_species):
    # Create genesdict
    gene2species= pd.DataFrame(
        list(df_BHs[[Query_accession,
                     Query_species]].values) + list(df_BHs[[Target_accession,
                                                            Target_species]].values
                                                   ),
        columns= ['gene', 'species']
    ).drop_duplicates().set_index('gene').species

    # Create species set
    species_list= sorted(set(gene2species))

    # Create map OG->species->gene
    OGs_table= pd.DataFrame([get_species_columns(genes, species_list, gene2species, idx)
                             for idx, genes in idx_2_og.items()],
                            columns= ['OG', 'n_genes', 'n_sepecies'] + species_list
                           )
    return OGs_table

def parse_mode(X):
    norm_factor, bidirectional= X.split('_')
    return colRelation[norm_factor], bidirectional=='bidirectional'

def parse_prt_files(path,
                    normalization_method,
                        Query_accession= 'Query_accession',
                        Target_accession= 'Target_accession',
                        Query_species= 'Query_species',
                        Target_species= 'Target_species',
                        pd_params= dict(names= _diamond_tl_headers, sep= '\t'),
                    species_parser= simple_species_parser,
                       ):
    print('Reading .proteinortho.tsv file and hits directory...')
    # Load prt hits
    df_hits= parse_prt_hits(norm_path(path),
                            pd_params= pd_params,
                            species_parser= species_parser,
                           )
    # Normalize if required
    if normalization_method in normalization_smallest:
        # (by the smallest sequences between target or query)
        _, bidirectional= parse_mode(normalization_method)
        df_hits['SMALLEST']= df_hits[['Query_length', 'Target_length']].min(axis=1)
        df_hits= normalize_scores(df_hits,
                                  Query_accession,
                                  Target_accession,
                                  'Bit_score',
                                  'SMALLEST',
                                  bidirectional,
                                  Query_species,
                                  Target_species)

    elif normalization_method in normalization_modes:
        # (By target, query, or aligment)
        norm_factor, bidirectional= parse_mode(normalization_method)
        df_hits= normalize_scores(df_hits,
                                  Query_accession,
                                  Target_accession,
                                  'Bit_score',
                                  norm_factor,
                                  bidirectional,
                                  Query_species,
                                  Target_species)
    return df_hits

def normalize_scores(df, a, b, score, norm_factor, bidirectional, a_s, b_s):
    normalized_score= pd.DataFrame({a_s : df[a_s],
                                    b_s : df[b_s],
                                    a : df[a],
                                    b : df[b],
                                    score : df[score]/df[norm_factor]})
    if bidirectional:
        DD= {(Q,T):S for Q,T,S in normalized_score[[a,b,score]].values}
        SS= dict(normalized_score[[a,a_s]].values)
        SS.update(dict(normalized_score[[b,b_s]].values))
        hits= {frozenset(X) for X in DD}
        normalize= lambda T,Q: (DD.get((Q,T), 0) + DD.get((T,Q), 0))/2
        D1= {(Q,T) : normalize(Q,T)
             for Q,T in hits}

        permutate= lambda X: (X, X[::-1])
        normalized_score= [(SS[Q],SS[T],Q,T,D1[X])
                           for X in D1 for Q,T in permutate(X)]
        normalized_score= pd.DataFrame(normalized_score,
                                       columns= [a_s,b_s,a,b,score]
                                      )

    return normalized_score    

def parse_prt_orthogroups(orthogroups_file, clear_singletones= True):
    df_prt= pd.read_csv(orthogroups_file, sep= '\t')
    species_cols= list(map(lambda x: x.split('.')[0], df_prt.columns[3:]))
    df_prt.columns= ['Species', 'Genes', 'Alg.-Conn.']+species_cols
    df_prt[species_cols]= df_prt[species_cols].apply(lambda X: X.apply(lambda x: set(x.split(','))))
    df_prt.index.name= 'OG'
    if clear_singletones:
        df_prt= _clean_prt(df_prt)
    return df_prt

def parse_prt_hits(hits_dir,
                   hit_file_ext= '.diamond',
                   Query_accession= 'Query_accession',
                   pd_params= dict(names= _diamond_tl_headers, sep= '\t'),
                   species_parser= simple_species_parser
                  ):
    """
    Returns a DataFrame of hits obtained from proteinortho files
    """

    df= pd.concat((_read_bh_table(f'{hits_dir}/{file}',
                                 pd_params,
                                 Query_accession= Query_accession,
                                 species_parser= species_parser,
                                 )
                   for file in os.listdir(hits_dir) if file.endswith(hit_file_ext)
                  ))
    return df

def _read_bh_table(file, pd_params, Query_accession, species_parser= simple_species_parser):
    """
    Para que funcione, el archivo debe ser llamado algo así como: 
    H0.vs.H10.diamond
    """
    species= species_parser(file)
    df= pd.read_csv(file, **pd_params)
    columns= list(df.columns)
    df['Query_species']= species[0]
    df['Target_species']= species[1]
    return df[ ['Query_species', 'Target_species']+columns ]

def max_hit_scores(df_hits,
                   Query_accession= 'Query_accession',
                   Target_species= 'Target_species',
                   Score= 'Bit_score',
                  ):
    return df_hits.groupby([Query_accession, Target_species])[Score].max()


def _get_orthogroup_hits(row, df_hits):
    """
    Input:
    - row: A row from an orthogroups table (.proteinortho.tsv)
    - df_hits: table of best hits (Query_gene -> Target_gene)
    """
    genes= set(chain.from_iterable(row))
    mask= (df_hits.Query_accession.isin(genes)) & (df_hits.Target_accession.isin(genes))
    og_hits= df_hits[ mask ]
    og_hits['OG']= row.name
    return og_hits

def _clean_prt(df_prt):
    mask= df_prt.Species > 1
    species= list(df_prt.columns[3:])
    return df_prt.loc[mask, species]

def read_symBets(prt_graph_path, prt_path):
    # Loads symBets (proteinortho-graph)
    symBets= pd.read_csv(prt_graph_path, sep='\t', comment= '#', names= ['a',
                                                                         'b',
                                                                         'evalue_ab',
                                                                         'bitscore_ab',
                                                                         'evalue_ba',
                                                                         'bitscore_ba',
                                                                        ])

    # Loads proteiortho table containng orthogroups
    df_prt= parse_prt_orthogroups(prt_path)

    # For each ortogroup obtain the orthology relations of all the genes in such orthogroup
    F= lambda x: _get_orthos(x, symBets)
    df_edges= pd.concat( list(df_prt.apply(F, axis= 1)) )

    return df_edges

def _get_orthos(row, symBets):
    orths= pd.concat((_search_orths(row, X, symBets) for X in row.index))
    orths['OG']= row.name
    return orths.set_index('OG').sort_index()

def _search_orths(row, species, symBets):
    mask= lambda x: symBets[ (symBets.a==x) | (symBets.b==x) ]
    return pd.concat(map(mask, row[species]))

####################
#                  #
# General functions#
#                  #
####################

def _check_files(path):
    prt_path= [x for x in os.listdir(path) if x.endswith('.proteinortho.tsv')]
    if len(prt_path)==1:
        prt_path= path + prt_path[0]
    else:
        raise ValueError(f'There is more than one ".proteinortho.tsv" file at {path}')
    return prt_path

def _get_prt_projectname(prt_file):
    return prt_file.split('/')[-1].split('.')[0]




import networkx as nx
from itertools import chain
from . import nxTree as nxt
from networkx.drawing.nx_pydot import graphviz_layout
from .error import InconsistentTrees

def BUILDst_N(profile, color_attr= None, *copyAttrs):
    """
    Input:
    NX trees con atributo color en los nodos
    Las hojas deben tener un label único dentro de cada árbol,
    pero deben de repetirse entre árboles.

    Para especificar el label, se define el kwordarg 'color_attr'. Por
    default el valor de este argumento es None.
    """
    # Init display graph
    H_p= _create_display_graph(profile, color_attr, copyAttrs)
    position= _get_initial_position(H_p)

    # Init output tree
    T= nx.DiGraph() # Create new tree
    T.add_node(0)
    T.root= 0
    nIdx= 1 # ID of the next inner node

    # Check if display graph is conected
    CC= list(nx.weakly_connected_components( _induced_sub_display(H_p, position) ))
    if len(CC) == 1:
        Q= [(position, T.root)]
    else:
        T.add_node(nIdx)
        T.add_edge(T.root, nIdx)
        Q= [(cc.intersection(position), nIdx) for cc in CC]
        nIdx+= 1

    while len(Q) > 0:
        position, pred= Q.pop(0)
        ind_leafs= _get_pseudoleafs(H_p, position, color_attr)
        len_ind_l= len(ind_leafs)
        if len_ind_l==1:
            leaf= ind_leafs.pop()
            attrs= H_p.nodes[leaf]
            T.add_node(nIdx, **attrs)
            T.add_edge(pred, nIdx)
            nIdx+= 1
        elif len_ind_l==2:
            leaf1, leaf2= ind_leafs
            XX= nIdx
            #
            T.add_node(XX)
            T.add_edge(pred, XX)
            nIdx+= 1
            #
            attrs= H_p.nodes[leaf1]
            T.add_node(nIdx, **attrs)
            T.add_edge(XX, nIdx)
            nIdx+= 1
            #
            attrs= H_p.nodes[leaf2]
            T.add_node(nIdx, **attrs)
            T.add_edge(XX, nIdx)
            nIdx+= 1
        else:
            T.add_node(nIdx)
            T.add_edge(pred, nIdx)
            position= _compute_successor(position, H_p)
            CC= list(nx.weakly_connected_components( _induced_sub_display(H_p, position) ))
            dQ= [(cc.intersection(position), nIdx) for cc in CC]
            if len(CC)==1:
                raise InconsistentTrees('Inconsistent profile')
            # Actualizar cola
            Q+= dQ
            nIdx+= 1

    return T

def _create_display_graph(profile, color_attr, copyAttrs):
    """
    Constructs a directed display graph [REvolutionH-tl paper]
    It is a modification of [display_graph paper]
    """
    H_p= nx.DiGraph()
    H_p.roots= set()
    nodes_relation, idx= _create_display_nodes(profile, H_p, color_attr, copyAttrs)
    _create_display_edges(profile, nodes_relation, H_p)
    _add_display_colors(H_p, color_attr, idx, copyAttrs)
    return H_p


def _create_display_nodes(profile, H_p, color_attr, copyAttrs):
    """
    Combine all the trees of in profile to a single directed graph H_p.

    Input:
    - Profile: list of nxTrees
    - H_p: empti nx.DiGraph with the attribute '.roots' as an empty set.
    - color_attr: string specifying the attribute for leafs color.

    This function manipulates the H_p graph as follows:
    - Addition of nodes: For each node v in the trees, it is created a node
      u in H_p.
    - Identification of roots: The roots of the original trees are stored
      in the attribute H_p.roots
    - Each of the created nodes have the attributes:
      - inner_label: Keeps the node v
      - tree_idx: Keeps the index in the profile where is stored de tree containing v
      - color: This attribute is copied from v

    Output:
    - nodes_relation (dict): A bijective mapping; for each tuple
      (T_i, v) there is a node u, where T_i is a tree in the profile,
      v is a vertex in T_i, and u is the corresponding vertex created
      in H_p.
    - idx (int): The total number of nodes in the original trees.
    """
    idx= 0
    nodes_relation= {}
    tree_idx= 0
    for X in profile:
        H_p.roots.add(idx)
        for x in nxt.BFS_graph(X, X.root):
            nodes_relation[(X,x)]= idx
            x_attrs= {'inner_label': x,
                      'tree_idx': tree_idx,
                     }

            if nxt.is_leaf(X, x):
                x_attrs.update({color_attr: X.nodes[x][color_attr]})
                x_attrs.update({atr: X.nodes[x][atr] for atr in copyAttrs})

            H_p.add_node(idx, **x_attrs)
            idx+= 1
        tree_idx+= 1
    return nodes_relation, idx

def _create_display_edges(profile, nodes_relation, H_p):
    """
    Add the edges in H_p that are equivalent to those in
    the profile using the map nodes_relation.
    """
    for X in profile:
        for x in X:
            idxx= nodes_relation[(X,x)]
            for y in X[x]:
                idxy= nodes_relation[(X,y)]
                H_p.add_edge(idxx, idxy)

def _add_display_colors(H_p, color_attr, idx, copyAttrs):
    """
    Create the a unique node for each of the colors present
    in the leafs of the profile, and creates an arc from such nodes
    to the leafs of the corresponding color.
    """
    colors= {}
    edges= []
    for x in H_p.roots:
        for y in nxt.induced_leafs(H_p, x):
            node_color= H_p.nodes[y][color_attr]
            if node_color in colors:
                ID= colors[node_color]
            else:
                colors[node_color]= ID= idx
                attrs= {color_attr: node_color}
                attrs.update({atr: H_p.nodes[y][atr] for atr in copyAttrs})
                H_p.add_node(ID, **attrs)
                idx+= 1
            H_p.add_edge(ID, y)
            H_p.nodes[y]['color_node']= ID
            

def _get_initial_position(H_p):
    return H_p.roots

def _get_pseudoleafs(H_p, position, color_attr):
    position_leafs= set(chain.from_iterable(nxt.induced_leafs(H_p, x) for x in position))
    #return {H_p.nodes[x][color_attr] for x in position_leafs}
    return {H_p.nodes[x]['color_node'] for x in position_leafs}

def _get_semiuniversal(position, H_p):
    tree_to_subPos= {}
    for x in position:
        tree_idx= H_p.nodes[x]['tree_idx']
        # Agregar el nodo a la lista de nodos del árbol
        # Si no existe dicha lista, la crea
        tree_to_subPos[ tree_idx ]= tree_to_subPos.get(tree_idx, []) + [x]
    return {L[0] for tree_idx,L in tree_to_subPos.items() if len(L)==1}

def _compute_successor(position, H_p):
    semiuniversal= _get_semiuniversal(position, H_p)
    semihijos= set(chain.from_iterable([set(H_p[x]) for x in semiuniversal]))
    return (position - semiuniversal).union(semihijos)

def _induced_sub_display(H_p, position):
    nbunch= {y for x in position for y in nxt.BFS_graph(H_p, x)}
    nbunch.update({H_p.nodes[x]['color_node'] for x in nbunch if nxt.is_leaf(H_p, x)})
    return nx.induced_subgraph(H_p, nbunch)




import pandas as pd
import networkx as nx
from itertools import chain
from tqdm import tqdm
tqdm.pandas()

_quey_target= ['Query', 'Target']

def tl_digraph_from_pandas(df, og_col= None):
    if og_col==None:
        og_col= 'OG'
        df[og_col]= 0

    df= df.set_index(og_col)
    return df.groupby(og_col).progress_apply(create_cBMG)

def read_tl_digraph(path, og_col= None):
    df= pd.read_csv(path, sep= '\t')
    return tl_digraph_from_pandas(df, og_col)

def write_digraphs_list(G, output_prefix= 'tl_project', n_attrs= ['accession', 'species'], mode= 'w', header= True):
    df= []
    for og, X in zip(G.index, G):
        df+= [ _get_edge_row(X, e, og, n_attrs) for e in X.out_edges ]
    df= pd.DataFrame(df, columns= get_digraph_cols(n_attrs))
    df.to_csv(output_prefix, sep= '\t', index= False, mode= mode, header= header)

def get_digraph_cols(n_attrs= ['accession', 'species']):
    return ['OG'] + [mod+'_'+attr for mod in _quey_target for attr in n_attrs]

def _get_edge_row(X, e, og, n_attrs):
    return [og] + [X.nodes[n][attr] for n in e for attr in n_attrs] #+ list(e)

def create_cBMG(df):
    try:
        og= df.index[0]
    except:
        print(df)
        1/0
    cBMG= nx.DiGraph()
    cBMG.og= og

    genes= set(df.Query_accession).union(df.Target_accession)
    DD= {x:i for i,x in enumerate(genes)}

    aux= chain.from_iterable((df[['Query_accession', 'Query_species']].drop_duplicates().values,
                              df[['Target_accession', 'Target_species']].drop_duplicates().values
                             ))
    aux= pd.DataFrame(list(aux)).drop_duplicates().values

    for gene,species in aux:
        cBMG.add_node(DD[gene],
                      species= species,
                      accession= gene,
                     )
       

    """
    df.apply(lambda row: cBMG.add_node(DD[row.Target_accession],
                                       species= row.Target_species,
                                       accession= row.Target_accession,
                                      ),
             axis= 1)
    """

    edges= df.apply(lambda row: (DD[row.Query_accession], DD[row.Target_accession]),
                    axis= 1
                   )

    cBMG.add_edges_from( edges )

    return cBMG




_df_matches_cols= ['Query_accession',
                  'Target_accession',
                        'Sequence_identity',
                        'Length',
                        'Mismatches',
                        'Gap_openings',
                        'Query_start',
                        'Query_end',
                        'Target_start',
                        'Target_end',
                        'E_value',
                        'Bit_score',
                       ]

_diamond_tl_headers= ['Query_accession',
                      'Target_accession',
                      'Query_length',
                      'Target_length',
                      'Alignment_ength',
                      'Bit_score',
                      'Evalue',
                     ]

_default_f= 0.95



import matplotlib.pyplot as plt
from networkx.drawing.nx_pydot import graphviz_layout
import numpy as np
import pydot
import networkx as nx
from .nxTree import is_leaf

def plot_tree(tree, label_attr= None, ax= None):
    if label_attr!=None:
        labels= {x: tree.nodes[x][label_attr] for x in tree}
    else:
        labels= None
    pos = graphviz_layout(tree, prog="dot", root= tree.root)
    nx.draw(tree,
            pos,
            with_labels= True,
            node_color= [tree.nodes[x]['color'] for x in tree],
            labels= labels,
            ax= ax,
           )
    return plt.gcf()

rgbColor= np.array([[102, 102, 255, 1], [255, 0, 102, 1], [102, 255, 153, 1]])/255
rgbColor[:,-1]= 1
colors_plot= {'red': rgbColor[0,:],
              'blue': rgbColor[1,:]
             }

def plot_digraph(G):
    pos= nx.circular_layout(G)

    nx.draw_networkx_nodes(G, pos, node_color= [colors_plot[G.nodes[x]['color']] for x in G])
    nx.draw_networkx_edges(G, pos,
                           connectionstyle="arc3,rad=0.15", edge_color= rgbColor[-1,:],
                           width= 3
                          )
    nx.draw_networkx_labels(G, pos)
    fig= plt.gcf()
    fig.set_size_inches(8,8)
    return fig




from .nxTree import copy_tree, is_leaf, induced_colors
from .nhxx_tools import read_nhxx, tralda_to_nxTree, get_nhx
from .in_out import get_digraph_cols, write_digraphs_list

from itertools import product, combinations
import networkx as nx
from bmgedit import BMGEditing
import pandas as pd
from math import ceil
from tqdm import tqdm
tqdm.pandas()


def build_graph(G,
                method='louvain',
                objective='cost',
                color_attr= None,
                use_binary_triples=True,
                force_binary= False,
                double_build= True,
               ):
    """
    ARG1: method - this param will be use to select the method
    """
    if color_attr != None:
        G1= G.copy()
        for x in G:
            G1.nodes[x]['color']= G.nodes[x][color_attr]
    else:
        G1= G

    solver = BMGEditing.BMGEditor(G1, binary= force_binary, use_binary_triples= use_binary_triples)
    solver.build(method, objective)

    hug_free= solver.get_bmg(extract_triples_first=False, update_tree= double_build)

    accession_2_species= {}
    nId_2_accession= {}
    for x in G:
        XX= G.nodes[x]['accession']
        YY= G.nodes[x]['species']
        hug_free.nodes[x]['accession']= XX
        hug_free.nodes[x]['species']= YY
        accession_2_species[XX]= YY
        nId_2_accession[x]= XX

    nxT= tralda_to_nxTree(solver._tree, name_attr= 'accession')
    # color_attr
    for x in nxT:
        if len(nxT[x])==0:
            nId= int(nxT.nodes[x].get('accession','NO_LABEL'))
            accession= nId_2_accession[nId]
            color= accession_2_species.get(accession)
            nxT.nodes[x]['species']= color
            nxT.nodes[x]['accession']= accession
        else:
            nxT.nodes[x]['species']= None

    return hug_free, nxT

def build_graph_series(G, args):
    lens= G.apply(len)
    is_sorted= all([lens.iloc[i] <= lens.iloc[i+1]
                    for i in range(lens.shape[0]-1)])
    if not is_sorted:
        order= argsort(lens.values)
        G= G.iloc[order]
        lens= lens.iloc[order]

    # Compute partition limits
    M= max(lens)
    f_low= lambda i: i*args.k_size_partition + 1
    f_up= lambda i: (i+1)*args.k_size_partition
    limits= ((f_low(i), f_up(i)) for i in range(0, ceil(M/args.k_size_partition)))
    total= len(G)
    Stotal= 0

    # File names
    bmg_opath= f'{args.output_prefix}.cBMGs.tsv'
    with open(bmg_opath, 'w') as F:
        F.write('\t'.join(get_digraph_cols())+'\n')
    ors_opath= f'{args.output_prefix}.orthologs.tsv'
    with open(ors_opath, 'w') as F:
        F.write('a\tb\tOG\n')
    gtr_opath= f'{args.output_prefix}.gene_trees.tsv'
    with open(gtr_opath, 'w') as F:
        F.write('OG\ttree\n')

    # Start process
    Tg_list= []
    for down,up in limits:
        mask= lens.apply(lambda x: down <= x <= up)
        lMask= sum(mask)

        if lMask>0:

            Stotal+= lMask
            print('\n')
            print(f'| Reconstructing evolution for families with {down} <= number of genes <= {up}')
            print(f'| This represents the {100*lMask/total:.2f} % of the data')
            print(f'| At the end of this bunch, {100*Stotal/total:.2f} % of the data will be analyzed.\n')
            print('Reconstructing gene trees...')

            G_T= G[mask].progress_apply(lambda X: build_graph(X,
                                                              color_attr= 'species',
                                                              method= args.bmg_heuristic,
                                                              use_binary_triples= args.no_binary_triples,
                                                              force_binary= args.force_binary_gene_tree,
                                                              double_build= args.gene_tree_no_double_build,
                                                             ))

            G[mask]= G_T.apply(lambda X: X[0])
            Tg= G_T.apply(lambda X: X[1])
            Tg.name='tree'

            print('Labeling gene tree with evolutionary events...')
            Tg= Tg.progress_apply( lambda x: get_augmented_tree(x, 'accession', 'species') )
            Tg_list+= [Tg]
            Tgtxt= Tg.apply(lambda x: get_nhx(x, root= 1, name_attr= 'accession')).reset_index()

            print('Writing best match graphs...')
            write_digraphs_list(G[mask], bmg_opath, mode= 'a', header= False)
            print(f'Successfully written to {bmg_opath}')
            print('Writing gene trees...')
            Tgtxt.to_csv(gtr_opath, sep='\t', index= False, mode= 'a', header= False)
            print(f'Successfully written to {gtr_opath} (This file will be used as input for step 3).')

    Tg= pd.concat(Tg_list)
    return Tg

##################################################################
##################################################################
##################################################################

def _addN(T,idx, tree, n):
    T.add_node(idx, **(tree.nodes[n].copy()))

def get_augmented_tree(tree, geneAttr, speciesAttr):
    """
    Constructs the color intersection graph
    Schaller et. al. 2020:
    Complete characterization of incorrect orthology assiments in BMGs
    """

    T= nx.DiGraph()
    T.root= 0
    _addN(T, 0, tree, 0)
    w_idx= 1
    mu= {1 : 0} # node x in tree --> corresponding dad in T

    for node in list(nx.dfs_preorder_nodes(tree, source= 1)):

        _addN(T, w_idx, tree, node)
        T.add_edge(mu[node], w_idx)
        for child in tree[node]:
            mu[child]= w_idx
        current= w_idx
        w_idx+= 1

        if (not is_leaf(tree, node)) and (node != tree.root):
            CIG= get_CIG(tree, node, speciesAttr)
            CC= list(nx.connected_components(CIG))
            if len(CC)>1:
                T.nodes[current][geneAttr]= 'S'
                for cc in (X for X in CC if len(X)>1):
                    _create_redundant_node(T, w_idx, current, tree, node, cc, mu, geneAttr, speciesAttr)
                    w_idx+= 1
            else:
                T.nodes[current][geneAttr]= 'D'
    T.nodes[T.root][geneAttr]= None
    return T

def get_CIG(tree, node, speciesAttr):
    """
    Color Intersection Graph
    """
    children= set(tree[node])
    CIG= nx.Graph()
    CIG.add_nodes_from(children)
    for v0,v1 in combinations(children, 2):
        colors0= induced_colors(tree, v0, speciesAttr)
        colors1= induced_colors(tree, v1, speciesAttr)
        if not colors0.isdisjoint(colors1):
            CIG.add_edge(v0, v1)
    return CIG

def _create_redundant_node(T, w_idx, current, tree, node, cc, mu, geneAttr, speciesAttr):
    T.add_node(w_idx, **{geneAttr: 'D', speciesAttr: None})
    T.add_edge(current, w_idx)
    for child in cc:
        mu[child]= w_idx



from .nxTree import induced_leafs, is_leaf
from revolutionhtl.triplets import get_triplets

import pandas as pd
from itertools import chain, combinations, product
from tqdm import tqdm
tqdm.pandas()

def get_orthologs(T, x= None, event= 'S', event_attr= 'event', label_attr= 'label', forbidden_nodes= [], forbidden_leaf_name= 'X'):
    """
    returns a list of lists, where each sublist is a pair of orthologous nodes
    i.e. leafs of the tree `T` such that their last common ancestor is an event of type `event`.
    """
    return _rec_orthologs(T,
                          forbidden_nodes,
                          forbidden_leaf_name,
                          x= x,
                          event= event,
                          event_attr= event_attr,
                          label_attr= label_attr)[1]

def orthologs_from_trees_df(df, forbidden_leaf_name= 'X', tree_col= 'gene_tree', orthogroup_column= 'OG'):
    df_orthologs= df.progress_apply(lambda row: _get_row_orth_table(row,
                                                                    tree_col,
                                                                    [],
                                                                    forbidden_leaf_name,
                                                                    'S',
                                                                    orthogroup_column,
                                                                   ), axis= 1)
    df_orthologs= pd.DataFrame(list(chain.from_iterable(df_orthologs)), columns= ['a', 'b', orthogroup_column])
    return df_orthologs

def _get_row_orth_table(row, tree_col, forbidden_nodes, forbidden_leaf_name, s_event, orthogroup_column):
    orths= get_orthologs(row[tree_col],
                         event= s_event,
                         event_attr= 'accession',
                         label_attr= 'accession',
                         forbidden_nodes= forbidden_nodes,
                         forbidden_leaf_name= forbidden_leaf_name
                        )
    return [list(X)+[row[orthogroup_column]] for X in orths]

def _rec_orthologs(T,
                   forbidden_nodes,
                   forbidden_leaf_name,
                   x= None,
                   event= 'S',
                   event_attr= 'event',
                   label_attr= 'label',
                  ):
    """
    Recursive function: a pos-order DFS traveling throug the nodes of 'T'.
    At each step it is returned the induced leaves and orthology relations induced
    by the node 'x'.
    """
    if x==None:
        x= T.root

    leaf_name= T.nodes[x][label_attr]
    if is_leaf(T, x) and leaf_name != forbidden_leaf_name:
        induced_leafs= [ leaf_name ]
        orths= []
    else:
        lefs_wrt_children= []
        orths= []
        for y in T[x]:
            y_leafs, y_orths= _rec_orthologs(T,
                                             forbidden_nodes,
                                             forbidden_leaf_name,
                                             x= y,
                                             event= event,
                                             event_attr= event_attr,
                                             label_attr= label_attr
                                            )
            lefs_wrt_children+= [y_leafs]
            orths+= y_orths
            
        induced_leafs= list(chain.from_iterable(lefs_wrt_children))

        if T.nodes[x][event_attr] == event and x not in forbidden_nodes:
            orths+= [set((a,b))
                     for A,B in combinations(lefs_wrt_children, 2)
                     for a,b in product(A,B)]

    return induced_leafs, orths

###############################
###############################
###############################

if __name__ == "__main__":
    import argparse
    from .nhxx_tools import read_nhxx

    parser = argparse.ArgumentParser(prog= 'revolutionhtl.orthology',
                                     description='Obtain orthology relations from a list of gene trees.',
                                    )
    # Parameters
    ############
    parser.add_argument('trees',
                        help= '.tsv file containing a gene tree in .nhx format (new newick) for each row in the column specified by -c',
                        type= str,
                       )

    parser.add_argument('-tc', '--tree_column',
                        help= 'column name for gene trees in the input .tsv file. (default="tree")',
                        default="tree",
                        type= str,
                       )

    parser.add_argument('-oc', '--orthogroup_column',
                        help= 'column name for gene trees ID in the input .tsv file. (default="OG")',
                        default="OG",
                        type= str,
                       )

    parser.add_argument('-l', '--loss_leafs',
                        help= 'Label of the leafs associated with a gene loss event. (default="X")',
                        default="X",
                        type= str,
                       )

    parser.add_argument('-o', '--output_prefix',
                    help= 'Prefix used for output files (default "tl_project").',
                    type= str,
                    default= 'tl_project',
                   )

    args= parser.parse_args()

    # Run program
    #############
    print('\n')
    print('revolutionhtl.orthology')
    print('----------------------------\n')

    # Read input
    print('Reading trees...')
    df_trees= pd.read_csv(args.trees, sep= '\t')
    df_trees[ args.tree_column ]= df_trees[ args.tree_column ].apply(lambda x: read_nhxx(x, name_attr= 'accession'))

    # Run
    print('Computing orthology...')
    df_orthologs= orthologs_from_trees_df(df_trees,
                                          forbidden_leaf_name= args.loss_leafs,
                                          tree_col= args.tree_column,
                                          orthogroup_column= args.orthogroup_column
                                         )

    # Write output
    print('Writing results...')
    opath= f'{args.output_prefix}.orthologs.tsv'
    df_orthologs.to_csv(opath, sep= '\t', index= False)
    print(f'Orthologs successfully written to {opath}')
#
